// required GameX definition and include:

#define GAMEX_MAIN
#pragma comment (lib, "gamex.lib")
#include "gamex.hpp"

// custom includes:

#include "sprite.hpp"
#include "resourcepool.hpp"
#include "text.hpp"

// global variables:

float screen_width, screen_height; // width and height of viewport, so we don't have to keep calling GameX.GetWidth() and GameX.GetHeight()
Sprite * main_char; // sprite of the main character (the escaping ship)
CameraX cam[2]; // two cameras used for 3D projection (cam[0] is for main view, cam[1] is for rear-view mirror (which caps lock turns on))

#define FIREBALL_NUM 16 // number of fireball sprites in the game
Sprite * fireball[FIREBALL_NUM]; // array of fireball sprites

#define STAR_NUM 64 // number of star dot sprites in the game
Sprite * star[STAR_NUM]; // array of star sprites

#define CHECKPOINT_DIST (50000.0f) // distance between each checkpoint ring

float camx, camy, camz; // global variables to keep track of main camera location
float vHealth = 0.0f; // health amount displayed on the health meter, moves toward actual health value over multiple cycles
int gameOver = 0; // 0 if the game is not over, 1 to 128 if the game is over
bool paused = false; // true if the game is paused, false if the game is not paused (the P key pauses the game)
int pauseFade = 0; // 0 to 128, determines how much to fade in the display of "Paused" over the screen
int score = -1; // current amount of points the player has accumulated

// custom function declarations:

void drawHealth (int number, float amount, float xCenter, float yCenter, float innerRadius, float outerRadius);
void drawCheckPoint (int ypos);

// required GameX functions:

// Required GameX initialize function (called when program starts)
// this function should initialize GameX and prepare anything else we need
void GameInit (void)
{
	// decide on a mode, which lets us request attributes for the game (see gamex-defines.hpp for more info)
	GameXInitFlags mode = VIDEO_16BIT|VIDEO_FULLSCREEN|RUN_ALWAYSHIDECURSOR;
	#ifdef _DEBUG
		mode |= RUN_AUTOSHOWINFO; // if in debug build, also have GameX show debugging info (frame rate, etc.)
	#endif

	// Initialize GameX itself (required):
	// "Ship Escape Demo" is the name of this game, it shows up in dialog boxes, the program name, etc.
	// 800 x 600 is the size we're requesting to make the viewport, note the standard 4:3 (width:height) pixel ratio
	GameX.Initialize ("Ship Escape Demo", mode, 800, 600);

	screen_width = (float)GameX.GetWidth();   // store the width and height of the viewport in global variables
	screen_height = (float)GameX.GetHeight(); // so we don't have to keep calling GameX.GetWidth() and GameX.GetHeight()

	// create the main character (ship) sprite with trail (jet) length 48, at the starting position
	main_char = new Sprite(SPRITE_MAIN_CHAR, 48, Vector3DF(0,200,0));

	// create each fireball sprite, with fireball type, a trail length of 64, and somewhat randomized position
	for(int i = 0 ; i < FIREBALL_NUM ; i++)
		fireball[i] = new Sprite(SPRITE_FIREBALL, 64, Vector3DF(GameX.GetRandomFloat(-2000,2000),300 -10000*(float)i,0));

	// create each star dot sprite, with star type, trail length of 0, and randomized position
	for(int j = 0 ; j < STAR_NUM ; j++)
		star[j] = new Sprite(SPRITE_STAR, 0, Vector3DF(GameX.GetRandomFloat(-2000,2000),99999,GameX.GetRandomFloat(-3000,3000)));

	// initialize the text printer (tells it where to find the alphabet graphic)
	tp.Initialize("Images\\text.gif");

	// initialize the resource pool (this function is customized to load all the graphics used by this game)
	rp.Initialize(); // (note that more complex games may need more sophisticated image management)

	// set up the cameras we'll use to project the scene
	cam[0].SetWindow(0,0,(int)screen_width,(int)screen_height); // tell the main camera to project to the whole screen
	cam[0].SetUpVector(Vector3DF(0,-1.0f,0)); // set -y to "up" so we can use a coordinate system like the standard 2D coordinate system (where +x means "right" and +y means "down") (note that +z means "in" in this system)
	cam[0].SetFarPlane(64000.0f); // set the maximum viewing distance to really far away so things don't suddenly appear when they get close
	cam[1].SetWindow(32,32,32+200,32+150); // tell the rear-view mirror camera to project to the upper-left corner
	cam[1].SetUpVector(Vector3DF(0,1.0f,0)); // this camera is looking backwards of the first camera
	cam[1].SetFarPlane(64000.0f); // again, reduce distance clipping by increasing maximum distance
	GameX.SetDrawCam(&cam[0]); // tell GameX what camera to draw with (the main one)

	// initialize some global variables that we'll use to keep track of where the main camera is located
	camx = 0; // horizontal (left/right) offset
	camy = 0; // vertical (up/down) offset
	camz = 768; // approximate distance above the scene (in/out offset)
}

void GameRun (void)
{
	// adjust the camera x and y positions to slowly "catch up" with the ship position
	camx = (camx*63.0f + main_char->position.x)/64.0f;
	camy = (camy*63.0f + main_char->position.y)/64.0f;

	// set the camera to be at its x,y,z positions
	cam[0].SetFromPosition(Vector3DF(camx,camy,-camz));

	// set the camera to look at the ship, offset slightly so it isn't exactly in the middle all the time
	cam[0].SetToPosition(Vector3DF(main_char->position.x/4,main_char->position.y-200,0));

	// NOTE: Because the x,y coords of the From position are not the same as those of the To position,
	//       the camera must look "up" into the distance to stay where it is and still look at its target.
	//
	//       If we wanted a completely overhead view, we would instead do:
	//			cam[0].SetFromPosition(main_char->position - Vector3DF(0,0,camz));
	//			cam[0].SetToPosition(main_char->position);

	// if the rear-view mirror is enabled, set the position of its camera, otherwise don't bother
	if(GameX.IsCapsLockOn() || GameX.IsKeyDown(KEY_R) || GameX.IsMouseDown(MOUSE_RIGHT)) {
		cam[1].SetFromPosition(Vector3DF(main_char->position.x,main_char->position.y,main_char->position.z+50));
		cam[1].SetToPosition(Vector3DF(main_char->position.x,main_char->position.y+1024,main_char->position.z));
	}

	// check for input that pauses the game:
	if(!gameOver) { // if the game is going,
		if(GameX.GetKeyPress(KEY_P)) { // check to see if the P key was JUST pressed.
		                               // note: using IsKeyDown or GetKeyDown would not work right here
			                           //       because then pressing P would register multiple times,
			                           //       causing rapid switching between paused and unpaused

			if((paused && pauseFade>48) || (!paused && !pauseFade)) // check to make slow-motion pause cheating less effective
				paused = !paused; // change paused to unpaused or vice-versa
		}
	}

	// if paused, don't advance anything non-visual in the game state
	if(paused) {
		if(pauseFade < 128) pauseFade++; // advance the pause fade amount (for purely visual purposes)
		return; // this return causes updates like ship position, collision detection, etc. to not happen now while paused
	} else {
		if(pauseFade > 0) pauseFade--; // reduce the pause fade amount (for purely visual purposes)
	}

	// remember previous y position so we can check whether the ship went through a ring after updating
	float yBefore = main_char->position.y;

	// tell the main character sprite to update itself (updates its position based on velocity, and updates its trail, etc.)
	main_char->update();

	// check to see if just passed through a checkpoint ring
	float yAfter = main_char->position.y;
	while(yBefore < 0) {
		yAfter += CHECKPOINT_DIST;
		yBefore += CHECKPOINT_DIST;
	}
	if(yAfter <= 0 && yBefore >= 0 && !gameOver) { // and if it did just pass through a ring,
		main_char->damage = max(0, main_char->damage - 10*100); // add 10% health
		main_char->damage = min(main_char->damage, (100-30)*100-1); // also restore to 30% minimum
		GameX.win_cps = 60; // turn off near-death slow motion effect (although it might have already been off)
		score += (int)(-100.0f*main_char->position.y/CHECKPOINT_DIST); // increase score
	}

	if(!gameOver) { // if the game is still going,
		            // check for and handle input:
		if(GameX.IsKeyDown(KEY_RIGHT))
			main_char->velocity.x += 1.0f; // the right arrow key increases the ship's rightward speed
		if(GameX.IsKeyDown(KEY_LEFT))
			main_char->velocity.x -= 1.0f; // the left arrow key increases the ship's leftward speed

		if(GameX.IsKeyDown(KEY_DOWN) && main_char->velocity.y <= -5.0f)
			main_char->velocity.y += 0.33f; // the down arrow key decreases the ship's upward speed if it's not already going too slow
		else {
			main_char->velocity.y -= 0.33f; // automatically increase the ships upward speed
			if(GameX.IsKeyDown(KEY_UP) && main_char->velocity.y > -30.0f)
				main_char->velocity.y -= 0.33f; // the up arrow key increases the ship's upward speed even more if it's not already going too fast
		}
		
		main_char->velocity.x += (float)GameX.GetMouseDX()/12.0f;

		float mouseDY = (float)GameX.GetMouseDY()/24.0f;
		if(mouseDY > 0 && main_char->velocity.y > -5.0f) mouseDY = 0; // prevent backtracking by mouse
		main_char->velocity.y += mouseDY;
	}

	// make sure the ship's speed stays within a reasonable range:
	// (these limits will probably only be reached when player really gets nailed)
	if(main_char->velocity.y < -50.0f)
		main_char->velocity.y = -50.0f;
	if(main_char->velocity.y > 25.0f)
		main_char->velocity.y = 25.0f;

	// if the ship is moving up fast enough, apply damping to its speed to limit it to about 30 units per 60th of a second
	if(main_char->velocity.y < -30.0f)
		main_char->velocity.y *= 0.95f;

	// apply a tiny bit of damping if it's above a somewhat lower upward, speed unless upward movement is being input
	if(main_char->velocity.y < -25.0f && !GameX.IsKeyDown(KEY_UP) && GameX.GetMouseDY() >= 0)
		main_char->velocity.y *= 0.99f;

	// apply horizontal damping so the ship's horizontal movement levels out fairly quickly
	main_char->velocity.x *= 0.95f;

	// for each fireball, aim it a bit more toward the ship and tell it to update itself
	for(int i = 0 ; i < FIREBALL_NUM ; i++) {
		if(!gameOver) {
			// attract the fireball toward the ship's position at a fairly high rate
			fireball[i]->attractTo(main_char->position, 0.425f);

			// if it's further "up" than the ship, make it home in more accurately:
			if(fireball[i]->position.y < main_char->position.y) {
				float ydist = fabsf(fireball[i]->position.y - main_char->position.y);
				float mx = main_char->position.x+main_char->velocity.x*ydist/60.0f;
				float gx = fireball[i]->position.x+fireball[i]->velocity.x*ydist/60.0f;
				if(mx > gx)
					fireball[i]->velocity.x += 0.33f;
				else if(mx < gx)
					fireball[i]->velocity.x -= 0.33f;
			}
		}

		fireball[i]->multiplyVelocity(0.99f); // drag
		fireball[i]->update(); // adds its current velocity to its current position
	}

	// now do collision detection:
	{
		for(int i = 0 ; i < FIREBALL_NUM ; i++) {
			for(int j = i+1 ; j < FIREBALL_NUM ; j++) {
				fireball[i]->collideCheck(fireball[j]); // between fireballs
			}
			fireball[i]->collideCheck(main_char); // between fireballs and the ship (or the ship's trail)
		}
	}

	if(gameOver) // if the game-over has been started,
		if(gameOver < 128) // if it hasn't advanced to the maximum we've picked
			gameOver++; // advance game-over status
		else
			GameX.Quit(true); // bad practice, faking fatal error to instantly quit...sometimes ok for demos or incomplete games
	else
		if(GameX.GetGlobalCounter()%60 == 0)
			score++; // increase the score every 60 cycles
		
	if(GameX.GetGlobalCounter()%16 == 0) { // only do these non-urgent calculations every 16 cycles to save a little CPU time:

		// check for slow-motion and game-over conditions based on the ship's health:
		float health = (float)(10000-main_char->damage)/10000.0f;
		health = max(health,0.0f);
		if(health < 0.1f) { // if health is below 10%
			float endFactor = health*10.0f; // calculate a factor for how drastic the slow-motion is
			GameX.win_cps = (int)(30+30*endFactor); // slow game time down to a minimum of 30 cycles per second as opposed to the normal 60 cps
		}
		if(!gameOver && health <= 0.0f) {
			gameOver = 1; // start the game-over
			paused = false; // just in case the game was somehow paused
		}

		if(!GameX.IsFullScreen()) // if the game is in windowed mode,
			GameX.SetMousePosition((int)screen_width/2,(int)screen_height/2); // prevent the mouse from (easily) leaving the window


		// move stars that are out of view to a new position in the distance
		// (so we can re-use a finite number of stars)
		{
			for(int i = 0 ; i < STAR_NUM ; i++) {
				if(star[i]->position.y > main_char->position.y+2000) {
					star[i]->position.y = main_char->position.y - GameX.GetRandomFloat(2000,10000);
				}
			}
		}
		
		// also recycle the fireballs by taking the one furthest behind and moving it forward into the distance
		// (only happens if there are no fireballs in front of the ship)
		{
			bool needNew = true;
			int low = -1;
			for(int i = 0 ; i < FIREBALL_NUM ; i++) {
				if(fireball[i]->position.y < main_char->position.y) {
					needNew = false;
					break;
				} else if(fireball[i]->position.y >= main_char->position.y+1000) {
					if(low == -1 || fireball[i]->position.y > fireball[low]->position.y)
						low = i;
				}
			}
			if(needNew && low != -1) {
				fireball[low]->position.y = main_char->position.y-10000;
			}
		}
	
	}
}

void GameDraw (void)
{
	// Calculate health percentage:
	float health = (float)(10000-main_char->damage)/10000.0f;
	vHealth = (vHealth*23.0f + health)/24.0f;
	int iHealth = (int)(health*100.0f);
	if(iHealth <= 0) {
		if(health > 0.0f) iHealth = 1;
		else              iHealth = 0;
	}

	// Clear the screen, but only clear it partway if health is low (for blur effect):
	int iHealthSq = (int)(health*health*100.0f);
	int alpha = (iHealthSq>32) ? 255 : iHealthSq*3+16;
	alpha -= (int)(1.5f*pauseFade);
	alpha = max(alpha, 16);
	GameX.SetDrawMode(DRAWOP_NODITHER);
	GameX.FillScreen(ColorX(0,0,0,alpha));


	// Render the game for each viewpoint (camera) we have
	// (the two viewpoints are: one for the main game view, one for the rear-view mirror.
	//  but another game might instead have one for each player's screen for a split-screen game)

	for(int viewpoint = 0 ; viewpoint < 2 ; viewpoint++) {

		if(viewpoint == 1) { // if this is the rear-view mirror viewpoint,

			if(!(GameX.IsCapsLockOn() || GameX.IsKeyDown(KEY_R) || GameX.IsMouseDown(MOUSE_RIGHT)))
				break; // skip drawing second viewpoint if the rear-view mirror isn't active

			// otherwise continue drawing and also draw a translucent frame first:
			GameX.DrawRect(ColorX(64,64,64,96),32-16,32-16,32+200+16,32+150+16);
			GameX.DrawRect(ColorX(0,0,0,128),32,32,32+200,32+150);
		}

		GameX.SetDrawCam(&cam[viewpoint]); // sets GameX to draw from the appropriate viewpoint

		// Now we draw the 3D scene. We should call Begin3DScene, then draw the scene, then call End3DScene.
		// However, we can optimize the drawing so the game runs smoother by moving certain transparent drawing
		// calls out of between the Begin and End 3D scene calls, more specifically, things that are numerous
		// and that we also know do not need to be drawn in any particular order we don't already know.
		// BUT, it is important to remember that 3D drawing outside of Begin3DScene and End3DScene is done
		// AT OUR OWN RISK OF POSSIBLE GLITCHES (be sure to test thoroughly if you do this!)

		GameX.Begin3DScene();
		// Because we just told GameX we're drawing a 3D scene,
		// the order of what we draw after this does not matter until the next call to End3DScene
		
		main_char->draw(); // draw the player's ship and its jet trail and/or explosion

		// Draw the star field (note: this would be moved to before Begin3DScene as an optimization if the starts were transparent)
		if(viewpoint != 1)
			for(int s = 0 ; s < STAR_NUM ; s++)
				star[s]->draw();

		// Draw the checkpoint rings (note: this loop and the use of constant checkpoint distance is terribly hack-ish)
		for(int checkPointY = 0 ; checkPointY > main_char->position.y - (CHECKPOINT_DIST + 1.0f) ; checkPointY -= (int)CHECKPOINT_DIST)
			if(main_char->position.y > checkPointY || checkPointY - main_char->position.y < 1024.0f)
				drawCheckPoint(checkPointY);

		GameX.End3DScene();

		// (Moved out of 3D scene calls as a drawing optimization, as described above.
		//  In this case the optimization is VERY effective because there are lots of pieces of fire being drawn
		//  for each fireball, and they are being drawn with DRAW_ADD which is commutative.)
		for(int i = 0 ; i < FIREBALL_NUM ; i++)
			fireball[i]->draw();

		// We are now done drawing the scene from this particular viewpoint.
	}

	// Now draw 2D graphics overlay such as health bar and text:
	//   (drawing order matters here!)

	drawHealth(iHealth, vHealth, screen_width-64, 64, 32.0f, 48.0f);

	if(gameOver > 0) {
		int c = gameOver*2;
		GameX.DrawRect(ColorX(0,0,0,c),0,0,(int)screen_width,(int)screen_height);
	}

	// Display the score (number of points the player has earned so far) in embossed text
	{
		int xCenter = (int)screen_width-128;
		int yCenter = 128-10;
		if(score > 9999) {
			xCenter -= (int)(10.0f * (float)((int)(log((double)(score/1000)) / log((double)10))));
		}
		char str [64];
		sprintf(str, "Score: %d", max(0,score));
		tp.PrintText(DRAW_ADD, 20.0f, str, xCenter-1, yCenter-1, 128,128,128);
		tp.PrintText(DRAW_SUBTRACT, 20.0f, str, xCenter+1, yCenter+1, 255,255,255);
		tp.PrintText(DRAW_ADD, 20.0f, str, xCenter-1, yCenter-1, 64+16,128+16,64+32);
	}

	if(GameX.GetGlobalCounter() < 1200) {
		tp.PrintText(DRAW_GHOST, 18.0f, "To dodge: Press the arrow keys or move the mouse.", 10, 10);
		tp.PrintText(DRAW_GHOST, 18.0f, "To see rear-view mirror: Press caps lock or hold R or right-click.", 10, 34);
		tp.PrintText(DRAW_GHOST, 18.0f, "To pause the game: Press the P key (it looks neat).", 10, 58);
		tp.PrintText(DRAW_GHOST, 18.0f, "To quit: Press Esc or Alt-F4 or Ctrl-Q (or lose).", 10, 82);
	}

	if(pauseFade > 0) {
		int c = pauseFade*2;
		int xCenter = (int)screen_width/2;
		int yCenter = (int)screen_height/2;
		tp.PrintText(DRAW_SUBTRACT|DRAWOP_CENTERED, 48.0f, "Paused", xCenter+1, yCenter+1, 128,128,128,c);
		tp.PrintText(DRAW_ADD|DRAWOP_CENTERED, 48.0f, "Paused", xCenter-1, yCenter-1, 128,128,128,c);

		if(paused && !GameX.IsFullScreen()) {
			// draw a simple cursor where the mouse is
			// if the game is paused and in windowed mode
			int x = GameX.GetMouseX();
			int y = GameX.GetMouseY();
			GameX.SetDrawScale(2.0f);
			GameX.SetDrawMode(DRAW_INVERT);
			GameX.DrawLine(ColorX(255,255,255,c),x-6,y,x+6,y);
			GameX.SetDrawScale(2.0f);
			GameX.SetDrawMode(DRAW_INVERT);
			GameX.DrawLine(ColorX(255,255,255,c),x,y-6,x,y+6);
		}
	}
}

void drawHealth (int number, float amount, float xCenter, float yCenter, float innerRadius, float outerRadius)
{
	float radius = (outerRadius + innerRadius) / 2.0f;
	float width = (outerRadius - innerRadius);

	int cp = (int)GameX.GetOscillatingValue(-16,16,1.0f/(amount+0.01f));
	float fullAng = 270.0f*DEGtoRAD;
	float ap = 3.0f*DEGtoRAD;
	float ae = 180.0f*DEGtoRAD;
	float maxAng = fullAng*amount + ap-0.01f;
	float xLast, yLast;

	// loop through angles, drawing lines in a circle around the center point to form the health bar:
	for(float angle = 0.0f ; angle <= fullAng ; ) {
		float amountHere = angle/fullAng;
	
		float x = radius*cosf(angle+ae);
		float y = radius*sinf(angle+ae);
		if(angle > 0.0f) {

			float xd = x-xLast;
			float yd = y-yLast;

			ColorX c2;
			ColorX c1 = ColorX(0,0,0,0);
			if(angle <= maxAng && number>0) {
				GameX.SetDrawScale(width*1.25f);
				int g = (int)(255.0f*sinf(amountHere*180.0f*DEGtoRAD));
				if(amountHere < 0.3f) g = max(g,128);
				c2 = ColorX((int)(3*cp+255*(1.0f-amountHere)),3*cp+g,(int)(3*cp+255*amountHere),255);
			} else {
				GameX.SetDrawScale(width*1.25f);
				c2 = ColorX(192,0,0);
			}
			ColorX c = ColorX(c1,c2,c1,c2);
			GameX.DrawLine(c,(int)(xLast+xCenter-xd-xd),(int)(yLast+yCenter-yd-yd),(int)(x+xCenter+xd+xd),(int)(y+yCenter+yd+yd));
		}
		xLast = x;
		yLast = y;

		if(fabsf(angle-maxAng) > ap+ap && fabsf(angle-fullAng) > ap+ap)
			angle += ap;
		else
			angle += DEGtoRAD;
	}
	
	int xc = (int)xCenter;
	int yc = (int)yCenter;

	// draw the bubble image in the middle of the health bar circle:
	GameX.SetDrawShading(ColorX(cp+128,cp+(int)(128.0f*sqrtf(amount)),cp+(int)(128.0f*sqrtf(amount))));
	GameX.SetDrawMode(DRAW_GHOST|DRAWOP_BRIGHT|DRAWOP_DITHER|DRAWOP_CENTERED);
	GameX.DrawImage(&rp.shellImage,xc,yc);

	// draw the player's shield percentage with embossed purplish text:

	char str [64];
	sprintf(str, "%d", number);
	tp.PrintText(DRAW_SUBTRACT|DRAWOP_CENTERED, 20.0f, str, xc+1, yc+1, 255,255,255);
	tp.PrintText(DRAW_ADD|DRAWOP_CENTERED, 20.0f, str, xc-1, yc-1, 128,160,192);

	tp.PrintText(DRAW_ADD, 16.0f, "Shields", xc-60-1, yc+64-35-1, 64,64,128);
	tp.PrintText(DRAW_SUBTRACT, 16.0f, "Shields", xc-60+1, yc+64-35+1, 128,128,128);
	tp.PrintText(DRAW_ADD, 16.0f, "Shields", xc-60-1, yc+64-35-1, 128,64,128);
}

void drawCheckPoint(int ypos)
{
	int segments = 16;
	float radius = 100.0f;
	float width = 32.0f;

	Vector3DF lastPoint = Vector3DF(0,0,0); // used since we (unnecessarily) draw each circle as a line-sprite

	float ap = 360.0f/(float)segments*DEGtoRAD;
	float ea = (float)GameX.GetGlobalCounter()/64.0f;
	float maxAng = 361.0f*DEGtoRAD;

	// loop through angles, drawing the bubble image at set intervals in a circle:
	for(float angle = 0.0f ; angle < maxAng ; angle += ap) {
		float x = radius * cosf(angle+ea);
		float z = radius * sinf(angle+ea);

		GameX.SetDrawScale(width);
		GameX.SetDrawMode(DRAW_GHOST|DRAWOP_CENTERED);

		// Set to draw with (smoothly transitioned) extra brightness if it's just been passed through:
		if(main_char->position.y <= ypos && ypos - main_char->position.y < 1024.0f) {
			float factor = 1.0f - (ypos - main_char->position.y) / 1024.0f;
			int c = (int)(256.0f*factor);
			GameX.SetDrawMode(DRAW_GHOST|DRAWOP_CENTERED|DRAWOP_BRIGHT);
			GameX.SetDrawShading(ColorX(c,c,c));
		}

		float closeness = fabsf(main_char->position.y - ypos)/1024.0f - 1.0f;
		float xpos;
		if(closeness < 1.0f)
			xpos = main_char->position.x;
		else if(closeness < 16.0f) {
			float closeFactor = (closeness-1.0f)/15.0f;
			xpos = camx*closeFactor + main_char->position.x*(1.0f-closeFactor);
		} else
			xpos = camx;

		Vector3DF thisPoint = Vector3DF(x+xpos,ypos-z,z+main_char->position.z);

		if(angle > 0) {
			Vector3DF v = thisPoint-lastPoint;
			GameX.DrawLineImage3D(&rp.shellImage,lastPoint-v,thisPoint+v,true);

			// alternatively, we could draw each bubble as a point-sprite instead of as a line-sprite,
			// allowing us to skip calculation of v and lastPoint,
			// by replacing the previous line of code with the following:
///			GameX.DrawImage3D(&rp.shellImage,thisPoint);
		}

		lastPoint = thisPoint;
	}
}